﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.UI;

public class JudgeManager : MonoBehaviour
{
    private static JudgeManager instance;
    public static JudgeManager Instance
    {
        get
        {
            if (instance == null)
            {
                instance = FindObjectOfType(typeof(JudgeManager))
                    as JudgeManager;
            }

            return instance;
        }
    }
    
    float latency = 225f;
    float scrollMultiplier = 1.0f;
    float startOffset = -5f;

    private GameObject offset;
    private GameObject noteobj, smobj, lmobj;
    private MotionGageManager motionGageManager;
    private GameObject judgeText;

    private GameObject motionSampleDisplayPrefab;

    private Vector3 initialPos;

    private float MsPerBeat
    {
        get
        {
            return 60 * 1000f / GameManager.Instance.CurrentTrack.BPM;
        }
    }
    
    private float ScrollSpeed
    {
        get
        {
            return scrollMultiplier * MsPerBeat / 1000f;
        }
    }

    void SetObjectRef()
    {
        offset = GameObject.Find("Offset");
        noteobj = GameObject.Find("Noteobj");
        smobj = GameObject.Find("SMobj");
        lmobj = GameObject.Find("LMobj");
        judgeText = GameObject.Find("Judge");
        motionGageManager = new MotionGageManager
        { motionGuage = GameObject.Find("Motion Guage") };
    }

    void JudgeInit()
    {
        initialPos = offset.transform.position;
        judgeText.SetActive(false);
        motionGageManager.ResetGuage();
    }

    void LoadGameObjects()
    {
        motionSampleDisplayPrefab = Resources.Load("Motion Sample Display") as GameObject;
        GameManager.Instance.CurrentTrack.Notes.ForEach(LoadNote);
        Instantiate(GameManager.Instance.defaultSound);
    }
    
    // Use this for initialization
    void Start()
    {
        SetObjectRef();
        JudgeInit();
        LoadGameObjects();
    }

    private bool IsNoteEnd()
    {
        return noteobj.transform.childCount <= 0 &&
                 smobj.transform.childCount <= 0 &&
                 lmobj.transform.childCount <= 0;
    }

    // Update is called once per frame
    void Update()
    {
        float timing = motionGageManager.UpdateTime(); ;

        offset.transform.position = Vector3.left * timing * ScrollSpeed;
        
        motionGageManager.UpdateGuage();

        if (IsNoteEnd())
            Invoke("ShowResult", 2f);

        ButtonNoteProc(timing);

        MotionNoteProc(timing);
    }

    void MotionNoteProc(float timing)
    {
        if (smobj.transform.childCount <= 0)
            return;

        GameObject smo = smobj.transform.GetChild(0).gameObject;
        MotionNote smnote = (MotionNote)smo.GetComponent<Note.Controller>().Instance;

        smnote.Checkpoint();

        if (!smnote.Activated && timing >= (smnote.StartTiming - MsPerBeat))
        {
            GameObject motionSample = Instantiate(motionSampleDisplayPrefab);
            MotionSampleDisplay msd = motionSample.GetComponent<MotionSampleDisplay>();
            msd.sprite = smnote.Image;
            msd.timeout = MsPerBeat;
            smnote.MotionSampleDisplay = msd;
            smnote.Activated = true;
        }
        if (timing >= (smnote.StartTiming - Judge.MaxButtonTimingRange))
        {
            if (smnote.FinalJudgeAction() || (timing > (smnote.EndTiming + Judge.MaxButtonTimingRange)))
            {
                SetJudge(Judge.TestJudge(smnote, timing, true), true);
                DeactivateNote(smnote);
            }
        }

        //            Debug.Log("T: " + timing + " nQ: " + activatedNotes.Count);

        /*
        for(int i=activatedNotes.Count-1;i>=0;i--)
        {
            MotionNote note = activatedNotes[i];
            if (note.FinalJudgeAction() || (timing > (note.EndTiming + Judge.MaxButtonTimingRange)))
            {
                SetJudge(Judge.TestJudge(note, timing + 350, false), true);
                activatedNotes.RemoveAt(i);
                DeactivateNote(note);
            }
        }*/

        /*
         if (lmobj.transform.childCount <= 0)
                return;

            GameObject lmo = lmobj.transform.GetChild(0).gameObject;
            MotionNote lmnote = (MotionNote)lmo.GetComponent<Note.Controller>().Instance;

            lmnote.Checkpoint();

            if (!lmnote.Activated && elapsedTime >= (lmnote.StartTiming - MsPerBeat))
            {
                GameObject motionSample = Instantiate(motionSampleDisplayPrefab);
                MotionSampleDisplay msd = motionSample.GetComponent<MotionSampleDisplay>();
                msd.sprite = lmnote.Image;
                msd.timeout = MsPerBeat;
                lmnote.Activated = true;
                lmnote.MotionSampleDisplay = msd;
                activatedNotes.Add(lmnote);
            }

            for (int i = activatedNotes.Count - 1; i >= 0; i--)
            {
                MotionNote note = activatedNotes[i];
                if (elapsedTime > note.StartTiming)
                    MotionGuageReset(note.Length);
                if (note.FinalJudgeAction() || elapsedTime > note.EndTiming + Judge.MaxButtonTimingRange)
                {
                    SetJudge(Judge.TestJudge(note, timing, true));
                    activatedNotes.RemoveAt(i);
                    DeactivateNote(note);

                    MotionGuageReset();
                }
            }
         */
    }

    void ButtonNoteProc(float timing)
    {
        if (noteobj.transform.childCount <= 0)
            return;

        GameObject obj = noteobj.transform.GetChild(0).gameObject;
        Note note = obj.GetComponent<Note.Controller>().Instance;

        if (IsRemainLongNoteProc(note, timing))
            return;

        Judge judge = Judge.TestJudge(note, timing);

        if (judge == Judge.MISS)
        {
            SetJudge(judge);
            DeactivateNote(note);
        }

        if (InputManager.Instance.ButtonStat == ButtonStatus.Pressed)
        {
            SetJudge(judge);

            if (judge == Judge.MISS) // Empty Miss
                return;

            if (note.IsLong)
                note.Activated = true;
            else
                DeactivateNote(note);
        }
    }

    bool IsRemainLongNoteProc(Note note, float timing)
    {
        if (!note.IsLong ||
            !note.Activated ||
            InputManager.Instance.ButtonStat == ButtonStatus.Pressed)
            return false;
        if (InputManager.Instance.ButtonStat == ButtonStatus.Released)
        {
            SetJudge(Judge.TestJudge(note, timing), true);
            DeactivateNote(note);
            return true;
        }

        if (Judge.TestJudge(note, timing, true) == Judge.MISS)
        {
            SetJudge(Judge.MISS);
            DeactivateNote(note);
        }

        return true;
    }

    private void DeactivateNote(Note note)
    {
        note.Activated = false;
        note.Component.transform.SetParent(offset.transform);
        note.Component.Deactivate();
    }

    private void SetJudge(Judge judge, bool isMotion = false)
    {
        if (!judge.IsBreak)
            ++GameManager.Instance.Combo;
        else
            GameManager.Instance.Combo = 0;
        GameManager.Instance.Score += judge.Score;
        GameManager.Instance.JudgeCount[judge]++;

        Debug.Log(judge.Name + (isMotion ? " Motion" : " Note") + " Combo: " + GameManager.Instance.Combo);

        judgeText.SetActive(true);
        judgeText.GetComponent<Text>().text = judge.Name;
        judgeText.GetComponent<Text>().color = judge.Color;
    }

    void LoadNote(Note note)
    {
        GameObject obj = Instantiate(
                Resources.Load(note.Type.ToString(), typeof(GameObject)),
                noteobj.transform)
                as GameObject;

        obj.AddComponent<Note.Controller>().Instance = note;

        SetNoteParent(note, obj);

        if (note.IsLong)
            StretchLongNote(note, obj);

        obj.transform.position += initialPos +
            (Vector3.right * (note.StartTiming * ScrollSpeed));
    }

    void SetNoteParent(Note note, GameObject obj)
    {
        if (note is MotionNote && !note.IsLong)
            obj.transform.SetParent(smobj.transform);
        else if (note is MotionNote && note.IsLong)
            obj.transform.SetParent(lmobj.transform);
        else if (note.Type == NoteType.MeasureLine || note.Type == NoteType.BeatLine)
            obj.transform.SetParent(offset.transform);
    }

    void StretchLongNote(Note note, GameObject obj)
    {
        float length = note.Length * ScrollSpeed;

        var holdTransform = obj.transform.GetChild(1)
            .GetComponent<RectTransform>();
        var endTransform = obj.transform.GetChild(2)
            .GetComponent<RectTransform>();

        holdTransform.SetSizeWithCurrentAnchors(
            RectTransform.Axis.Horizontal, length);
        holdTransform.position += (Vector3.right * length / 2);
        endTransform.position += (Vector3.right * length);
    }

    private bool onResult = false;

    private void ShowResult()
    {
        if (!onResult)
        {
            GameManager.Instance.SceneTransition("Result", true);
            onResult = true;
        }
    }
}

internal class MotionGageManager : MonoBehaviour
{
    public GameObject motionGuage;
    private float elapsedMotion;
    private float motionTimeout;
    private float elapsedTime = 0;

    public void ResetGuage(float timeout = 0f)
    {
        if (timeout <= 0f)
        {
            elapsedMotion = 0f;
            motionGuage.transform.parent.gameObject.SetActive(false);
            return;
        }

        motionTimeout = timeout;
        motionGuage.transform.parent.gameObject.SetActive(true);
        motionGuage.SetActive(true);
    }
    public void UpdateGuage()
    {
        if (motionGuage.transform.parent.gameObject.activeInHierarchy)
            UpdateTime();

        if (elapsedMotion >= motionTimeout)
            ResetGuage();
        else
            motionGuage.GetComponent<Image>().fillAmount = elapsedMotion / motionTimeout;
    }

    public float UpdateTime()
    {
        elapsedTime += Time.deltaTime * 1000;

        return elapsedTime;
    }
}